/*
 * xmms-mad - mp3 plugin for xmms
 * Copyright (C) 2001-2002 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: xmms-mad.c,v 1.15 2005/02/10 14:06:23 sam Exp $
 */
#include "config.h"
#include "xmms-mad.h"
#include "input.h"

#include <gtk/gtk.h>
#include <audacious/util.h>
#include <audacious/configdb.h>
#include <stdarg.h>
#include <fcntl.h>
#include <audacious/vfs.h>

/*
 * Global variables
 */
struct xmmsmad_config_t xmmsmad_config;   /**< global configuration */
static GStaticMutex mutex;

/*
 * static variables
 */
static GThread *decode_thread; /**< the single decoder thread */
static struct mad_info_t info;   /**< info for current track */

#ifndef NOGUI
static GtkWidget *about_window = 0;
static GtkWidget *error_dialog = 0;
#endif /* !NOGUI */

extern gboolean scan_file (struct mad_info_t *info, gboolean fast);

static void
xmmsmad_init ()
{
  ConfigDb *db = NULL;
  xmmsmad_config.http_buffer_size = HTTP_BUFFER_SIZE,
  xmmsmad_config.fast_play_time_calc = TRUE;
  xmmsmad_config.use_xing = TRUE;
  xmmsmad_config.dither = TRUE;
  db = bmp_cfg_db_open();
  if(db)
    {
      bmp_cfg_db_get_int(db, "MAD", "http_buffer_size", &xmmsmad_config.http_buffer_size);
      bmp_cfg_db_get_bool(db, "MAD", "fast_play_time_calc", &xmmsmad_config.fast_play_time_calc);
      bmp_cfg_db_get_bool(db, "MAD", "use_xing", &xmmsmad_config.use_xing);
      bmp_cfg_db_get_bool(db, "MAD", "dither", &xmmsmad_config.dither);
      bmp_cfg_db_close(db);
    }
  g_static_mutex_init(&mutex);
}

static void
xmmsmad_cleanup ()
{
}

static void
xmmsmad_set_eq (int on, float preamp, float *bands)
{
  //printf("set_eq\n");
}

static int
xmmsmad_is_our_file (char *filename)
{
  int fin = 0;
  int rtn = 0;
  guchar check [4];

  if (strncasecmp ("http://", filename, 7) == 0)
    {
      if (strcasecmp ("mp3", filename + strlen (filename) - 3) == 0)
        {
          /* [FIXME] do some checks here to make sure we have an mp3 stream */
          rtn = 1;
        }
    }
  else
    {
      fin = open (filename, O_RDONLY);
      if (fin >= 0 && read (fin, check, 4) == 4)
	{
	  /* If first two bytes are a sync header or three bytes are "ID3" */
	  if ( (check[0] == 0xff && (check[1] & 0x70) == 0x70) 
	       || memcmp (check, "ID3", 3) == 0) 
	    {
	      rtn = 1;
	    }
	  else if (memcmp (check, "RIFF", 4) == 0) /* RIFF data */
	    {        
	      lseek (fin, 4, SEEK_CUR);
	      read (fin, check, 4);

	      if (memcmp (check, "RMP3", 4) == 0)
		{
		  rtn = 1;
		}
#if 0
	      /* FIXME: make this test endian safe */
	      else if (memcmp (check, "WAVE"))
	        {
		  /* seek to codec ID and see if it's MP3 */
		  lseek (fin, 20, SEEK_SET);
		  check = 0L;
		  read (fin, &check, 2);
		  if (check == 0x55)
		    {
		      rtn = 1;
		    }
		}
#endif 
	    }
	}
    }

  close (fin);
  return rtn;
}

static void
xmmsmad_stop ()
{
#ifdef DEBUG
  g_message ("f: xmmsmad_stop");
#endif /* DEBUG */
  g_static_mutex_lock (&mutex);
  if (decode_thread)
    {
	    info.stop=1;
#ifdef DEBUG
      g_message ("waiting for thread");
#endif /* DEBUG */
      g_thread_join(decode_thread);
#ifdef DEBUG
      g_message ("thread done");
#endif /* DEBUG */
      input_term (&info);
      decode_thread = NULL;
      info.stop = 0;
    }
  g_static_mutex_unlock (&mutex);
}


static void
xmmsmad_play_file (char *url)
{
  gboolean rtn;

  xmmsmad_stop ();

#ifdef DEBUG
  g_message ("playing %s", url);
#endif /* DEBUG */

  if (input_init (&info, url) == FALSE)
    {
      g_message ("error initialising input");
      return;
    }

  rtn = input_get_info (&info, xmmsmad_config.fast_play_time_calc);

  if (rtn == FALSE)
    {
      g_message ("error reading input info");
      return;
    }
  decode_thread = g_thread_create(decode, (void *) &info, TRUE, NULL);
}

static void
xmmsmad_pause (short paused)
{
  mad_plugin.output->pause (paused);
}

static void
xmmsmad_seek (int time)
{
  /* xmms gives us the desired seek time in seconds */
  info.seek = time;
}

static int
xmmsmad_get_time ()
{
  if (info.stop)
    {
#ifdef DEBUG
      g_message ("returning -1 from get_time");
#endif
      return -1;
    }
  return mad_plugin.output->output_time ();
}

/**
 * Scan the given file or URL.
 * Fills in the title string and the track length in milliseconds.
 */
//static void
void
xmmsmad_get_song_info (char *url, char **title, int *length)
{
  struct mad_info_t myinfo;
#ifdef DEBUG
  g_message ("f: xmmsmad_get_song_info: %s", url);
#endif /* DEBUG */
  input_init (&myinfo, url);
  if (input_get_info (&myinfo, TRUE) == TRUE)
    {
      *title  = strdup (myinfo.title);
      *length = mad_timer_count (myinfo.duration, MAD_UNITS_MILLISECONDS);
    }
  else
    {
      *title = strdup (url);
      *length = -1;
    }
  input_term (&myinfo);
#ifdef DEBUG
  g_message ("e: xmmsmad_get_song_info");
#endif /* DEBUG */
}

static void
xmmsmad_about ()
{
    static GtkWidget *aboutbox;

    if (aboutbox != NULL)
        return;

    aboutbox = xmms_show_message("About MAD MPEG Decoder Plugin",
                                 "Based on libmad",
                                 "Ok",
                                 FALSE, NULL, NULL);

    g_signal_connect(G_OBJECT(aboutbox), "destroy",
                     G_CALLBACK(gtk_widget_destroyed), &aboutbox);
}



/**
 * Display a GTK box containing the given error message.
 * Taken from mpg123 plugin.
 */
void
xmmsmad_error (char *error, ...)
{
#ifndef NOGUI
  if (!error_dialog)
    {
      va_list args;
      char string[256];
      va_start (args, error);
      vsnprintf (string, 256, error, args);
      va_end (args);
      GDK_THREADS_ENTER ();
      error_dialog = xmms_show_message ("Error", string, "Ok", FALSE, 0, 0);
      gtk_signal_connect (GTK_OBJECT (error_dialog),
			  "destroy",
			  GTK_SIGNAL_FUNC (gtk_widget_destroyed),
			  &error_dialog);
      GDK_THREADS_LEAVE ();
    }
#endif /* !NOGUI */
}

extern void xmmsmad_get_file_info (char *filename);
extern void xmmsmad_configure ();


/*
 * Function extname (filename)
 *
 *    Return pointer within filename to its extenstion, or NULL if
 *    filename has no extension.
 *
 */
static gchar *
extname(const char *filename)
{
    gchar *ext = strrchr(filename, '.');

    if (ext != NULL)
        ++ext;

    return ext;
}

// tuple stuff
static TitleInput *
get_song_tuple(char *filename)
{
  TitleInput *tuple = NULL;
  gchar *string = NULL;
  VFSFile *file;

  struct id3_file *id3file = NULL;
  struct id3_tag  *tag = NULL;
  struct id3_frame *frame = NULL;
  union id3_field *field;

#ifdef DEBUG
  string = str_to_utf8(filename);
  g_message ("f: mad: get_song_tuple: %s", string);
  if(string){
	  g_free(string);
	  string = NULL;
  }
#endif

  if((file = vfs_fopen(filename, "rb")) != NULL) {
	  tuple = bmp_title_input_new();

	  id3file = id3_file_open (filename, ID3_FILE_MODE_READONLY);

	  if (id3file) {
		  tag = id3_file_tag (id3file);

		  if (tag) {
			  tuple->performer = input_id3_get_string (tag, ID3_FRAME_ARTIST);
			  tuple->album_name = input_id3_get_string (tag, ID3_FRAME_ALBUM);
			  tuple->track_name = input_id3_get_string (tag, ID3_FRAME_TITLE);

			  // year
			  string = NULL;
			  string = input_id3_get_string (tag, ID3_FRAME_YEAR); //TDRC
			  if(!string)
				  string = input_id3_get_string (tag, "TYER");

			  if(string){
				  tuple->year = atoi(string);
				  g_free(string);
				  string = NULL;
			  }

			  tuple->file_name = g_path_get_basename(filename);
			  tuple->file_path = g_path_get_dirname(filename);
			  tuple->file_ext = extname(filename);

			  // length
			  {
				  char *dummy = NULL;
				  int length;
				  xmmsmad_get_song_info(filename, &dummy, &length);
				  tuple->length = length;
				  if(dummy)
					  g_free(dummy);
			  }
				  
			  // track number
			  string = input_id3_get_string (tag, ID3_FRAME_TRACK);
			  if (string) 
			  {
				  tuple->track_number = atoi (string);
				  g_free (string);
				  string = NULL;
			  }

			  // genre
			  tuple->genre = input_id3_get_string (tag, ID3_FRAME_GENRE);

			  // comment
			  tuple->comment = input_id3_get_string (tag, ID3_FRAME_COMMENT);

		  }
		  id3_file_close(id3file);
	  }
	  vfs_fclose(file);
  }
#ifdef DEBUG
  g_message ("e: mad: get_song_tuple");
#endif
  return tuple;

}



/**
 * Retrieve meta-information about URL.
 * For local files this means ID3 tag etc.
 */
gboolean
mad_get_info (struct mad_info_t *info, gboolean fast_scan)
{
	TitleInput *tuple = NULL;

  g_message ("f: mad_get_info: %s", info->filename);

#if 0
  if (info->remote)
    { /* remote stream */
      if (input_http_init (info))
        return FALSE;
      if (input_udp_init (info))
        return FALSE;
    }
  else
#endif
    { /* local mp3 file */
	tuple = get_song_tuple(info->filename);
	info->title = xmms_get_titlestring (xmms_get_gentitle_format (), tuple);
      /* scan mp3 file, decoding headers unless fast_scan is set */
      if (scan_file (info, fast_scan) == FALSE)
	{
	  return FALSE;
	}
      /* reset the input file to the start */
      lseek (info->infile, 0, SEEK_SET);
      info->offset = 0;
      /* use the filename for the title as a last resort */
      if (!info->title) 
	{
		char *pos = strrchr (info->filename, '/');
	  if (pos)
	    info->title = g_strdup (pos + 1);
	  else
	    info->title = g_strdup (info->filename);
	}
    }
  g_message ("e: mad_get_info");
  return TRUE;
}

/**
 * XMMS input plugin hook.
 */

InputPlugin mad_plugin = {
	NULL,
	NULL,
	NULL,
	xmmsmad_init,
	xmmsmad_about,
	xmmsmad_configure,
	xmmsmad_is_our_file,
	NULL,
	xmmsmad_play_file,
	xmmsmad_stop,
	xmmsmad_pause,
	xmmsmad_seek,
	xmmsmad_set_eq,
	xmmsmad_get_time,
	NULL,
	NULL,
	xmmsmad_cleanup,
	NULL,
	NULL,
	NULL,
	NULL,
	xmmsmad_get_song_info,
	xmmsmad_get_file_info,
	NULL,
	get_song_tuple,	//get song tuple
	NULL,	//set song tuple
	NULL  // status buffering
};

InputPlugin *
get_iplugin_info(void)
{
	mad_plugin.description = DESCRIPTION " " VERSION;
	return &mad_plugin;
}
